# (c) 2014, Maciej Delmanowski <drybjed@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

import netaddr


# ---- IP address and network filters ----

def ipaddr(value, query = '', version = False, alias = 'ipaddr'):
    ''' Check if string is an IP address or network and filter it '''

    query_types = [ 'type', 'bool', 'int', 'version', 'size', 'address', 'ip', 'host', \
                    'network', 'subnet', 'prefix', 'broadcast', 'netmask', 'hostmask', \
                    'unicast', 'multicast', 'private', 'public', 'loopback', 'lo', \
                    'revdns', 'wrap', 'ipv6', 'v6', 'ipv4', 'v4', 'cidr', 'net', \
                    'hostnet', 'router', 'gateway', 'gw', 'host/prefix', 'address/prefix' ]

    if not value:
        return False

    elif value == True:
        return False

    # Check if value is a list and parse each element
    elif isinstance(value, (list, tuple)):

        _ret = []
        for element in value:
            if ipaddr(element, str(query), version):
                    _ret.append(ipaddr(element, str(query), version))

        if _ret:
            return _ret
        else:
            return list()

    # Check if value is a number and convert it to an IP address
    elif str(value).isdigit():

        # We don't know what IP version to assume, so let's check IPv4 first,
        # then IPv6
        try:
            if ((not version) or (version and version == 4)):
                v = netaddr.IPNetwork('0.0.0.0/0')
                v.value = int(value)
                v.prefixlen = 32
            elif version and version == 6:
                v = netaddr.IPNetwork('::/0')
                v.value = int(value)
                v.prefixlen = 128

        # IPv4 didn't work the first time, so it definitely has to be IPv6
        except:
            try:
                v = netaddr.IPNetwork('::/0')
                v.value = int(value)
                v.prefixlen = 128

            # The value is too big for IPv6. Are you a nanobot?
            except:
                return False

        # We got an IP address, let's mark it as such
        value = str(v)
        vtype = 'address'

    # value has not been recognized, check if it's a valid IP string
    else:
        try:
            v = netaddr.IPNetwork(value)

            # value is a valid IP string, check if user specified
            # CIDR prefix or just an IP address, this will indicate default
            # output format
            try:
                address, prefix = value.split('/')
                vtype = 'network'
            except:
                vtype = 'address'

        # value hasn't been recognized, maybe it's a numerical CIDR?
        except:
            try:
                address, prefix = value.split('/')
                address.isdigit()
                address = int(address)
                prefix.isdigit()
                prefix = int(prefix)

            # It's not numerical CIDR, give up
            except:
                return False

            # It is something, so let's try and build a CIDR from the parts
            try:
                v = netaddr.IPNetwork('0.0.0.0/0')
                v.value = address
                v.prefixlen = prefix

            # It's not a valid IPv4 CIDR
            except:
                try:
                    v = netaddr.IPNetwork('::/0')
                    v.value = address
                    v.prefixlen = prefix

                # It's not a valid IPv6 CIDR. Give up.
                except:
                    return False

            # We have a valid CIDR, so let's write it in correct format
            value = str(v)
            vtype = 'network'

    v_ip = netaddr.IPAddress(str(v.ip))

    # We have a query string but it's not in the known query types. Check if
    # that string is a valid subnet, if so, we can check later if given IP
    # address/network is inside that specific subnet
    try:
        if query and query not in query_types and ipaddr(query, 'network'):
            iplist = netaddr.IPSet([netaddr.IPNetwork(query)])
            query = 'cidr_lookup'
    except:
        None

    # This code checks if value maches the IP version the user wants, ie. if
    # it's any version ("ipaddr()"), IPv4 ("ipv4()") or IPv6 ("ipv6()")
    # If version does not match, return False
    if version and v.version != version:
        return False

    # We don't have any query to process, so just check what type the user
    # expects, and return the IP address in a correct format
    if not query:
        if v:
            if vtype == 'address':
                return str(v.ip)
            elif vtype == 'network':
                return str(v)

    elif query == 'type':
        if v.size == 1:
            return 'address'
        if v.size > 1:
            if v.ip != v.network:
                return 'address'
            else:
                return 'network'

    elif query == 'bool':
        if v:
            return True

    elif query == 'int':
        if vtype == 'address':
            return int(v.ip)
        elif vtype == 'network':
            return str(int(v.ip)) + '/' + str(int(v.prefixlen))

    elif query == 'version':
        return v.version

    elif query == 'size':
        return v.size

    elif query in [ 'address', 'ip' ]:
        if v.size == 1:
            return str(v.ip)
        if v.size > 1:
            if v.ip != v.network:
                return str(v.ip)

    elif query == 'host':
        if v.size == 1:
            return str(v)
        elif v.size > 1:
            if v.ip != v.network:
                return str(v.ip) + '/' + str(v.prefixlen)

    elif query == 'net':
        if v.size > 1:
            if v.ip == v.network:
                return str(v.network) + '/' + str(v.prefixlen)

    elif query in [ 'hostnet', 'router', 'gateway', 'gw', 'host/prefix', 'address/prefix' ]:
        if v.size > 1:
            if v.ip != v.network:
                return str(v.ip) + '/' + str(v.prefixlen)

    elif query == 'network':
        if v.size > 1:
            return str(v.network)

    elif query == 'subnet':
        return str(v.cidr)

    elif query == 'cidr':
        return str(v)

    elif query == 'prefix':
        return int(v.prefixlen)

    elif query == 'broadcast':
        if v.size > 1:
            return str(v.broadcast)

    elif query == 'netmask':
        if v.size > 1:
            return str(v.netmask)

    elif query == 'hostmask':
        return str(v.hostmask)

    elif query == 'unicast':
        if v.is_unicast():
            return value

    elif query == 'multicast':
        if v.is_multicast():
            return value

    elif query == 'link-local':
        if v.version == 4:
            if ipaddr(str(v_ip), '169.254.0.0/24'):
                return value

        elif v.version == 6:
            if ipaddr(str(v_ip), 'fe80::/10'):
                return value

    elif query == 'private':
        if v.is_private():
            return value

    elif query == 'public':
        if v_ip.is_unicast() and not v_ip.is_private() and \
            not v_ip.is_loopback() and not v_ip.is_netmask() and \
            not v_ip.is_hostmask():
            return value

    elif query in [ 'loopback', 'lo' ]:
        if v_ip.is_loopback():
            return value

    elif query == 'revdns':
            return v_ip.reverse_dns

    elif query == 'wrap':
        if v.version == 6:
            if vtype == 'address':
                return '[' + str(v.ip) + ']'
            elif vtype == 'network':
                return '[' + str(v.ip) + ']/' + str(v.prefixlen)
        else:
            return value

    elif query in [ 'ipv6', 'v6' ]:
        if v.version == 4:
            return str(v.ipv6())
        else:
            return value

    elif query in [ 'ipv4', 'v4' ]:
        if v.version == 6:
            try:
                return str(v.ipv4())
            except:
                return False
        else:
            return value

    elif query == '6to4':

        if v.version == 4:

            if v.size == 1:
                ipconv = str(v.ip)
            elif v.size > 1:
                if v.ip != v.network:
                    ipconv = str(v.ip)
                else:
                    ipconv = False

            if ipaddr(ipconv, 'public'):
                numbers = list(map(int, ipconv.split('.')))

            try:
                return '2002:{:02x}{:02x}:{:02x}{:02x}::1/48'.format(*numbers)
            except:
                return False

        elif v.version == 6:
            if vtype == 'address':
                if ipaddr(str(v), '2002::/16'):
                    return value
            elif vtype == 'network':
                if v.ip != v.network:
                    if ipaddr(str(v.ip), '2002::/16'):
                        return value
                else:
                    return False

    elif query == 'cidr_lookup':
        try:
            if v in iplist:
                return value
        except:
            return False

    else:
        try:
            float(query)
            if v.size == 1:
                if vtype == 'address':
                    return str(v.ip)
                elif vtype == 'network':
                    return str(v)

            elif v.size > 1:
                try:
                    return str(v[query]) + '/' + str(v.prefixlen)
                except:
                    return False

            else:
                return value

        except:
            raise errors.AnsibleFilterError(alias + ': unknown filter type: %s' % query)

    return False


def ipwrap(value, query = ''):
    try:
        if isinstance(value, (list, tuple)):
            _ret = []
            for element in value:
                if ipaddr(element, query, version = False, alias = 'ipwrap'):
                    _ret.append(ipaddr(element, 'wrap'))
                else:
                    _ret.append(element)

            return _ret
        else:
            _ret = ipaddr(value, query, version = False, alias = 'ipwrap')
            if _ret:
                return ipaddr(_ret, 'wrap')
            else:
                return value

    except:
        return value


def ipv4(value, query = ''):
    return ipaddr(value, query, version = 4, alias = 'ipv4')


def ipv6(value, query = ''):
    return ipaddr(value, query, version = 6, alias = 'ipv6')


# Split given subnet into smaller subnets or find out the biggest subnet of
# a given IP address with given CIDR prefix
# Usage:
#
#  - address or address/prefix | ipsubnet
#      returns CIDR subnet of a given input
#
#  - address/prefix | ipsubnet(cidr)
#      returns number of possible subnets for given CIDR prefix
#
#  - address/prefix | ipsubnet(cidr, index)
#      returns new subnet with given CIDR prefix
#
#  - address | ipsubnet(cidr)
#      returns biggest subnet with given CIDR prefix that address belongs to
#
#  - address | ipsubnet(cidr, index)
#      returns next indexed subnet which contains given address
def ipsubnet(value, query = '', index = 'x'):
    ''' Manipulate IPv4/IPv6 subnets '''

    try:
        vtype = ipaddr(value, 'type')
        if vtype == 'address':
            v = ipaddr(value, 'cidr')
        elif vtype == 'network':
            v = ipaddr(value, 'subnet')

        value = netaddr.IPNetwork(v)
    except:
        return False

    if not query:
        return str(value)

    elif str(query).isdigit():
        vsize = ipaddr(v, 'size')
        query = int(query)

        try:
            float(index)
            index = int(index)

            if vsize > 1:
                try:
                    return str(list(value.subnet(query))[index])
                except:
                    return False

            elif vsize == 1:
                try:
                    return str(value.supernet(query)[index])
                except:
                    return False

        except:
            if vsize > 1:
                try:
                    return str(len(list(value.subnet(query))))
                except:
                    return False

            elif vsize == 1:
                try:
                    return str(value.supernet(query)[0])
                except:
                    return False

    return False


# ---- HWaddr / MAC address filters ----

def hwaddr(value, query = '', alias = 'hwaddr'):
    ''' Check if string is a HW/MAC address and filter it '''

    try:
        v = netaddr.EUI(value)
    except:
        if query and query not in [ 'bool' ]:
            raise errors.AnsibleFilterError(alias + ': not a hardware address: %s' % value)

    if not query:
        if v:
            return value

    elif query == 'bool':
        if v:
            return True

    elif query in [ 'win', 'eui48' ]:
        v.dialect = netaddr.mac_eui48
        return str(v)

    elif query == 'unix':
        v.dialect = netaddr.mac_unix
        return str(v)

    elif query in [ 'pgsql', 'postgresql', 'psql' ]:
        v.dialect = netaddr.mac_pgsql
        return str(v)

    elif query == 'cisco':
        v.dialect = netaddr.mac_cisco
        return str(v)

    elif query == 'bare':
        v.dialect = netaddr.mac_bare
        return str(v)

    elif query == 'linux':
        v.dialect = mac_linux
        return str(v)

    else:
        raise errors.AnsibleFilterError(alias + ': unknown filter type: %s' % query)

    return False

class mac_linux(netaddr.mac_unix): pass
mac_linux.word_fmt = '%.2x'


def macaddr(value, query = ''):
    return hwaddr(value, query, alias = 'macaddr')


# ---- Ansible filters ----

class FilterModule(object):
    ''' IP address and network manipulation filters '''

    def filters(self):
        return {

            # IP addresses and networks
            'ipaddr': ipaddr,
            'ipwrap': ipwrap,
            'ipv4': ipv4,
            'ipv6': ipv6,
            'ipsubnet': ipsubnet,

            # MAC / HW addresses
            'hwaddr': hwaddr,
            'macaddr': macaddr

        }




# deepops.py
#
# Copyright (C) Robert Franklin <rcf@mince.net>



"""Deep operations module.

This module contains functions to perform operations on hierarchical
structures of dictionaries, lists and sets.
"""



__version__ = "1.1 (2017-09-26)"



# --- internal functions ---



def _strquote(s):
    """Returns the string representation of an object, surrounding it
    with single quotes, if it's a string (to make it clear it's a
    string and enclose spaces).

    This module uses it in various exception messages.

    Keywork arguments:

    s -- the item to be returned, string or other
    """

    return ("'%s'" % s) if isinstance(s, str) else s



# --- external functions ---



# deepmerge



def _deepmerge(a, b, replace, list_as_set, path):
    """Backend function for deepmerge() that does the actual work.  It
    is defined privately to not offer the 'path' argument.

    See deepmerge() for information.

    Keyword arguments (in addition to deepmerge()):

    path -- this is a list of keys identifying the position in the
    hierarchy of dictionaries being merged: it is used when raising an
    exception to explain where the problem lies
    """


    # if the items being merged are both lists, what we do depends on
    # the list_as_set option...

    if isinstance(a, list) and isinstance(b, list):
        # ... if it's enabled, we treat the list 'a' as a set and only
        # add items from 'b' to it if they don't exist already

        if list_as_set:
            for item in b:
                if item not in a:
                    a.append(item)


        # ... or, if it's disabled, we just append the corresponding
        # list in 'b' to 'a' (potentially adding duplicates)

        else:
            a.extend(b)


    # if the items being merged are both sets, we just add the missing
    # items in 'b'

    elif isinstance(a, set) and isinstance(b, set):
        a.update(b)


    # if the items being merged are both dictionaries, we work through the
    # items in 'b' in see if they're present in 'a'...

    elif isinstance(a, dict) and isinstance(b, dict):
        for item in b:
            if item in a:
                # the item is present - what we do now depends on the type of
                # that object (we only check 'a')...

                if (isinstance(a[item], dict) or
                    isinstance(a[item], list) or
                    isinstance(a[item], set)):

                    # the item is a compound type (dictionary, list or set) -
                    # recursively merge them
                    #
                    # we don't need to check if the two items are different
                    # types as the function will do that itself

                    _deepmerge(
                        a[item], b[item], replace, list_as_set, path + [item])


                else:
                    # the item in 'a' is a simple type - check the
                    # corresponding item in 'b' is also simple and replace it
                    # (if that's enabled), otherwise raise an exception

                    if (isinstance(b[item], dict) or
                        isinstance(b[item], list) or
                        isinstance(b[item], set)):

                        raise(TypeError("cannot replace simple type %s with "
                                        "non-simple type %s at %s" %
                                            (_strquote(a[item]),
                                             _strquote(b[item]),
                                             ".".join(path + [item]))))

                    if replace:
                        a[item] = b[item]


            else:
                # the item exists in 'b' but not in 'a', so just add the item
                # to 'a'

                a[item] = b[item]


    else:
        raise(TypeError("merging incompatible or illegal types %s and %s at "
                        "%s" % (_strquote(a), _strquote(b),
                                ".".join(path) if path else ".")))



def deepmerge(a, b, replace=True, list_as_set=False):
    """Merge two dictionaries - 'a' and 'b' - the items in 'b' are
    merged into 'a', in place, modifying 'a'.

    What happens for any particular item depends on its type:

    For lists, the corresponding items in 'b' are appended.  This will
    create duplicates, if they're in both lists.

    For sets, the result is the union of the items in 'a' and 'b'.

    For simple types (integers, strings, etc.), the value in 'b'
    replaces the value in 'a'.

    For dictionaries (including the toplevel call), the merge is
    applied recursively, with items in 'b' are added to 'a' by merging
    recursively.  Other types are merged according to the above rules.

    If one of the values is a simple type and the other is a compound
    type (list or dictionary) then a TypeError() exception is raised.

    Keyword arguments:

    a -- the 'initial' dictionary: this dictionary will be modified in
    place to contain the merged items from 'b'

    b -- the 'additional' dictionary: items in this dictionary will be
    added to 'a', taking precedence over them

    replace -- this indicates whether a simple type in 'a' should be
    replaced by a value in 'b'; this effectively determines whether
    values in 'a' or 'b' have precedence

    list_as_set -- this specifies whether lists should be treated as
    lists (= False), resulting in [potentially] duplicated items, or as
    sets (= True), which will only add items from 'b' if they don't
    already exist in 'a' (although note that it won't filter out
    duplicates already in 'a')
    """

    _deepmerge(a, b, replace, list_as_set, [])



# deepremoveitems



def _deepremoveitems(a, b, path):
    """Backend function for deepremoveitems() that does the actual
    work.  It is defined privately to not offer the 'path' argument.

    See deepremoveitems() for information.

    Keyword arguments (in addition to deepremoveitems()):

    path -- this is a list of keys identifying the position in the
    hierarchy of dictionaries being processed: it is used when raising
    an exception to explain where the problem lies
    """


    # we cannot remove a simple type (one that is not a dictionary, list or
    # set), regardless of the type of object we're removing them from (we'll
    # check for that later), so just abort with an exception

    if not(isinstance(b, dict) or isinstance(b, list) or isinstance(b, set)):
        raise(TypeError("cannot remove simple type %s at %s" %
                            (_strquote(b), ".".join(path) if path else ".")))


    # if the object we're removing from is a list or set...

    if isinstance(a, list) or isinstance(a, set):
        # ... and the object specifying what to remove is also a list or set,
        # we just remove any items that are in the removal list

        if isinstance(b, list) or isinstance(b, set):
            for item in b:
                if item in a:
                    a.remove(item)


        # ... or, if the object specifying what to remove is a dictionary, we
        # remove any items matching the keys of the dictionary, as long as the
        # dictionary is empty
        #
        # if the dictionary is not empty, we raise an exception as that implies
        # want to remove specific items from another dictionary (we have to
        # allow empty dictionaries to provide a way to remove a complete item
        # from the dictionary)

        elif isinstance(b, dict):
            for item in b:
                if not(b[item]):
                    if item in a:
                        a.remove(item)
                else:
                    raise(ValueError("cannot remove non-empty dictionary item "
                                     "%s from non-dictionary %s at %s" %
                                         (_strquote(item), _strquote(a),
                                          ".".join(path) if path else ".")))


    # if the object we're removing from is a dictionary...

    elif isinstance(a, dict):
        # ... and the object specifying what to remove is a list or set, we
        # just remove the keys in that list or set, if they exist

        if isinstance(b, list) or isinstance(b, set):
            for item in b:
                if item in a:
                    a.pop(item)


        # ... or, if the object specifying what to remove is also a dictionary,
        # we either remove the entire key, if the corresponding dictionary item
        # is empty
        #
        # if the item is not empty, we recursively process the two dictionaries
        # to remove the corresponding items

        elif isinstance(b, dict):
            for item in b:
                if item in a:
                    if not(b[item]):
                        a.pop(item)
                    else:
                        _deepremoveitems(a[item], b[item], path + [item])


    # if the object we're removing from is not one of the above - probably a
    # simple type - we're raise an exception as that's not supported

    else:
        raise(TypeError("cannot remove items from simple type %s at %s" %
                            (a, ".".join(path) if path else ".")))


def deepremoveitems(a, b):
    """Remove items 'b' from a object 'a', modifying object 'a' in
    place.  Both 'a' and 'b' must be compound types (a dictionary, list
    or set).

    What happens depends on the types of corresponding items in 'a' and
    'b':

    Where 'a' is a list or set, the items are removed, if they are
    present in 'b'.  'b' can be a list, set or dictionary but, if the
    corresponding item in 'b' is a dictionary, it must be empty.

    Where 'a' is a dictionary, 'b' can either be a list or set (in
    which case, the items in 'b' are removed from 'a'), or a
    dictionary.  In the latter case, if the corresponding item in 'b'
    is empty, the item in 'a' is removed entirely; if not, the items in
    each dictionary are processed recursively.

    The reason for handling empty dictionaries as a special case is to
    allow a mix of removing entire items (which could be done with a
    list or set for 'b') and removing specific items within other items
    (which requires a dictionary).  Note that an empty list or set
    cannot be used instead of an empty dictionary: they will remove
    no items from the corresponding item.

    If there are mismatches, violating the above rules, a TypeError()
    or ValueError() exception is raised.

    Keyword arguments:

    a -- the object to have items removed from it: this can be a
    dictionary, list or set and will be modified in place to remove the
    corresponding items in 'b'

    b -- the object specifying what items are to be removed from 'a';
    it doesn't need to match the type of 'a' but there are certain
    constraints, described above, which it must fit within, at each
    level
    """

    _deepremoveitems(a, b, [])



# net_genconfig.__main__


import net_genconfig

import argparse
import copy
import datetime
import jinja2
import os
import re
import sys
import yaml

from net_genconfig import deepops
from net_genconfig import netaddr_filter



# --- functions ---



def raise_exception(e):
    """Simple function to raise the specified exception.  This is used
    for the parameter onerror for os.walk(), which otherwise ignores
    errors and skips over problematic directory entries.
    """

    raise e


def warn_helper(msg):
    """Helper function to print a warning inside a Jinja2 template.  The
    supplied message is printed to stderr but execution is not stopped.

    Keyword arguments:

    msg -- warning message to display
    """

    print("warning: " + msg, file=sys.stderr)


    # helper functions return a string to be included in the template; we don't
    # particularly have anything to include but have to return something, so we
    # return an empty string

    return ""


def raise_helper(msg):
    """Helper function to raise an exception inside a Jinja2 template.  The
    generic 'Exception' class is raised in Python.

    Keyword arguments:

    msg -- exception message to display when aborting
    """

    raise Exception(msg)


def assert_helper(condition, msg):
    """Helper function for assertions inside a Jinja2 template.  If the
    supplied condition is False, an exception is raised with the given
    error message.  The 'AssertionError' class is raised by Python.

    If the condition is True, an empty string is returned to avoid
    printing anything.

    Keyword arguments:

    condition -- the condition which must be satisfied to not raise the
    exception

    msg -- message to display in the event of the condition not being
    true
    """

    if not condition:
        raise AssertionError(msg)

    return ""


def re_match_helper(pattern, string):
    """Helper function for matching strings against regular expressions
    in a Jinja2 template.  The regular expression is compiled and
    matched against the supplied string (the Python regular expression
    library caches compiled strings for efficiency) and a boolean is
    returned, indicating whether the match succeeded or not.

    Keyword arguments:

    pattern -- the regular expression to match against

    string -- the string to match against
    """

    return bool(re.match(pattern, string))


def re_match_groups_helper(pattern, string):
    """Helper function for matching strings against regular expressions
    and extract groups of (separated by parentheses) in a Jinja2
    template.  The regular expression is compiled and matched against
    the supplied string (the Python regular expression library caches
    compiled strings for efficiency) and the groups in the expression
    are returned in the form of a list.

    Keyword arguments:

    pattern -- the regular expression to match against; the required
    groups to return should be specified using parentheses

    string -- the string to match and return the groups of
    """

    r = re.match(pattern, string)

    if r is None:
        raise ValueError("'%s' does not match regular expression '%s'" %
                             (string, pattern))

    return r.groups()



# --- command line arguments ---



# create the parser and add in the available command line options

parser = argparse.ArgumentParser(
    # override the program name as running this as a __main__ inside a module
    # directory will use '__main__' by default - this name isn't necessarily
    # correct, but it looks better than that
    prog="net-genconfig",

    # we want the epilog help output to be printed as it and not reformatted or
    # line wrapped
    formatter_class=argparse.RawDescriptionHelpFormatter)


parser.add_argument(
    "-t", "--templates",
    default=((os.environ.get("NET_CONFIG_DIR") + "/templates")
                 if "NET_CONFIG_DIR" in os.environ else "templates"),
    help="directory containing configuration library templates")

parser.add_argument(
    "-i", "--inventory",
    default=((os.environ.get("NET_CONFIG_DIR") + "/inventory")
                 if "NET_CONFIG_DIR" in os.environ else "inventory"),
    help="directory containing inventory of devices, networks, etc.")

parser.add_argument(
    "-a", "--add-templates",
    action="append",
    default=[],
    help="additional directories in which to search for templates")

parser.add_argument(
    "-d", "--define",
    action="append",
    nargs=2,
    default=[],
    help="define variable for use in the template",
    metavar=("VAR", "VALUE"))

parser.add_argument(
    "-D", "--debug",
    action="store_true",
    help="enable debug mode")

parser.add_argument(
    "devicename",
    help="name of the device to generate the configuration for")

parser.add_argument(
    "--version",
    action="version",
    version=("%(prog)s " + net_genconfig.__version__))


# parse the supplied command line against these options, storing the results

args = parser.parse_args()

templates_dir = args.templates
inventory_dir = args.inventory
add_templates_dirs = args.add_templates
devicename = args.devicename
debug = args.debug

vars = {}
for var, value in args.define:
    vars[var] = value



# --- inventory ---



if debug:
    print("debug: starting to read inventory", file=sys.stderr)



# trawl the inventory directory tree, read and merge all the YAML files into
# one big dictionary

inventory = dict()

item_filepaths = dict()

try:
    # walk the directory tree, raising an exception if there are any
    # problems (oddly, os.walk() doesn't do this by default and skips
    # over any errors)

    for dirpath, dirnames, filenames in \
        os.walk(inventory_dir, onerror=raise_exception):

        # remove any directories whose name begins with a dot (to skip over
        # things like '.git'; '.' and '..' are already removed) so they're not
        # traversed into on subsequent iterations

        for dirname in dirnames:
            if dirname.startswith("."):
                dirnames.remove(dirname)


        # work through all the files in this directory, in sorted order, and
        # process them

        for filename in sorted(filenames):
            # skip files beginning with a dot (assumed to be temporary,
            # control, or other unwanted files)

            if filename.startswith("."):
                continue


            # read this file as YAML and add it to the inventory

            filepath = os.path.join(dirpath, filename)

            if debug:
                print("debug: reading inventory file '%s'" % filepath,
                      file=sys.stderr)

            try:
                inventory_file = yaml.safe_load(open(filepath))

            except ValueError as exception:
                print("error: failed parsing inventory file: '%s': %s" %
                          (filepath, exception),
                      file=sys.stderr)

                exit(1)


            for area in inventory_file:
                # create this part of the inventory, if it doesn't exist

                if area not in inventory:
                    inventory[area] = dict()
                    item_filepaths[area] = dict()


                else:
                    # if it did exist, raise an error if we already have a
                    # sub-item with the same name in that part

                    for item in inventory_file[area]:
                        if item in inventory[area]:
                            print("error: duplicate entry in inventory "
                                  "files: %s.%s in '%s' and '%s'" %
                                      (area, item, filepath,
                                       item_filepaths[area][item]),
                                  file=sys.stderr)

                            exit(1)


                # add the items to the (possibly empty) part of the inventory

                inventory[area].update(inventory_file[area])


                # record where we found those items

                for item in inventory_file[area]:
                    item_filepaths[area][item] = filepath


                if debug:
                    print("debug: read %d %s" % (len(inventory_file), area),
                          file=sys.stderr)


except OSError as e:
    # this exception catches subclasses of OSError, including things
    # like 'file not found' or 'permission denied'

    print("error: problem reading inventory:", e, file=sys.stderr)

    exit(1)



if debug:
    print("debug: finished reading inventory", file=sys.stderr)



if "devices" not in inventory:
    print("error: no devices found in inventory", file=sys.stderr)
    exit(1)


if devicename not in inventory["devices"]:
    print("error: device not found in inventory: ", devicename,
          file=sys.stderr)

    exit(1)



# get the definition dictionary for this device from the inventory; because
# we're not going to modify it, we don't need to copy it
#
# we use this variable for some checks here, and to pass it to the starting
# role template, so it avoids repeatedly fetching it
#
# if we're using a device profile, we will overwrite this after merging the
# profiles in

device = inventory["devices"][devicename]


if device is None:
    print("error: definition for device '%s' is empty" % devicename,
          file=sys.stderr)

    exit(1)


# if we have profiles for this device, we need to use those as the base for the
# device definition, so we need to merge them all together and into the
# explicit device definition itself

if "profiles" in device:
    if "device-profiles" not in inventory:
        print("error: profile '%s' used for device '%s' but no device "
              "profiles in inventory" % (device["profile"], devicename),

              file=sys.stderr)

        exit(1)


    # this dictionary will be used to assemble and merge the list of profiles
    # referenced that we can merge the device configuration into

    profiles_merged = {}


    # go through the imported device profiles in order

    for profile in device["profiles"]:
        # check this particular profile is defined

        if profile not in inventory["device-profiles"]:
            print("error: profile '%s' not found for device '%s'"
                      % (profile, devicename),

                  file=sys.stderr)

            exit(1)


        # it exists - merge a deep copy of the profile into the merged profile
        #
        # we need to deep copy the profiles as we may remove things from them
        # due to the 'profiles-exclude' option

        deepops.deepmerge(profiles_merged,
                          copy.deepcopy(inventory["device-profiles"][profile]))


    # if we're excluding anything from the imported profiles, we now remove
    # these from the merged profile we just assembled

    if "profiles-exclude" in inventory["devices"][devicename]:
        deepops.deepremoveitems(profiles_merged, device["profiles-exclude"])


    # finally, we merge the device definition dictionary over the top which
    # will, by default, replace any clashing items, favouring the ones in the
    # local device definition
    #
    # we then replace the device definition with the merged one

    deepops.deepmerge(profiles_merged, device)

    device = profiles_merged



if "role" not in device:
    print("error: missing role for device '%s'" % devicename,
          file=sys.stderr)

    exit(1)


if "platform" not in device:
    print("error: missing platform for device '%s'" % devicename,
          file=sys.stderr)

    exit(1)


role = device["role"]
platform = device["platform"]



# --- generate configuration ---



if debug:
    print("debug: generating configuration for '%s' with role '%s' running "
          "platform '%s'" % (devicename, role, platform), file=sys.stderr)


if not os.path.isdir(templates_dir):
    print("error: template directory does not exist: '%s'" % templates_dir,
          file=sys.stderr)

    exit(1)


# build the Jinja2 environment

env = jinja2.Environment(
    loader=jinja2.FileSystemLoader([templates_dir] + add_templates_dirs),
    extensions=[
        "jinja2.ext.do", "jinja2.ext.loopcontrols", "jinja2.ext.with_"],
    trim_blocks=True)


# add in the special warn(), raise() and assert() functions, as well as some
# other functions we need, into the Jinja2 environment

env.globals["warn"] = warn_helper
env.globals["raise"] = raise_helper
env.globals["assert"] = assert_helper
env.globals["re_match"] = re_match_helper
env.globals["re_match_groups"] = re_match_groups_helper

env.globals["deepcopy"] = copy.deepcopy

env.globals["deepmerge"] = deepops.deepmerge
env.globals["deepremoveitems"] = deepops.deepremoveitems


# add in the netaddr library functions as additional filters

for filter_name, filter_func in (
    netaddr_filter.FilterModule().filters().items()):

    env.filters[filter_name] = filter_func


if not os.path.isdir("%s/roles/%s" % (templates_dir, platform)):
    print("error: platform not found: '%s'" % platform, file=sys.stderr)
    exit(1)


# read the template file

template_file = "roles/%s/%s.j2" % (platform, role)

if debug:
    print("debug: using template file '%s'" % template_file, file=sys.stderr)

try:
    template = env.get_template(template_file, ".")

except jinja2.exceptions.TemplateNotFound:
    print("error: role not found: '%s' for platform '%s'" % (role, platform),
          file=sys.stderr)

    exit(1)


# render the template and print it

output = template.render(devicename=devicename, device=device,
                         inventory=inventory, **vars)

print("""\
!
! === %s (role %s / platform %s) ===
! Configuration generated %s
!


""" %
    (devicename.lower(), role, platform,
     datetime.date.strftime(datetime.datetime.now(), "%Y-%m-%d %H:%M:%S")))

print(output)



#!/usr/bin/env python3


import setuptools

import net_genconfig


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="net-genconfig",
    version=net_genconfig.__version__,
    author="Robert Franklin",
    author_email="rcf34@cam.ac.uk",
    description="Network device configuration generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.uis.cam.ac.uk/",
    packages=setuptools.find_packages(),
    install_requires=[
        "jinja2",
        "netaddr",
        "pyyaml",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
